from copy import deepcopy as copy
import numpy as np


class BmView:
    def __init__(self, data=None):
        self.bms = []
        self.op_seed_list = []
        self.y_all_list = []
        self.y_opt_list = []
        self.t_list = []
        self.is_init = False

        if data is not None:
            self.init_from_data(data)

    @property
    def is_group(self):
        return len(self.bms) > 0

    @property
    def t_best(self):
        return np.min(self.t_list)

    @property
    def t_mean(self):
        return np.mean(self.t_list)

    @property
    def t_wrst(self):
        return np.max(self.t_list)

    @property
    def y_all_best(self):
        y_lists = [self.values_to_opt(y_all) for y_all in self.y_all_list]
        return self.values_to_one(y_lists, kind='best')

    @property
    def y_all_mean(self):
        y_lists = [self.values_to_opt(y_all) for y_all in self.y_all_list]
        return self.values_to_one(y_lists, kind='mean')

    @property
    def y_all_wrst(self):
        y_lists = [self.values_to_opt(y_all) for y_all in self.y_all_list]
        return self.values_to_one(y_lists, kind='wrst')

    @property
    def y_opt_best(self):
        y = self.y_opt_list
        return np.max(y) if self.is_max else np.min(y)

    @property
    def y_opt_mean(self):
        return np.mean(self.y_opt_list)

    @property
    def y_opt_wrst(self):
        y = self.y_opt_list
        return np.min(y) if self.is_max else np.max(y)

    def add(self, bm):
        self.bms.append(bm)
        if len(self.bms) == 1:
            self.init_from_bm(bm)

        self.op_seed_list.append(bm.op_seed)
        self.y_all_list.append(bm.y_list)
        self.y_opt_list.append(bm.y_opt)
        self.t_list.append(bm.t)
        if bm.is_fail:
            self.is_fail = True

    def get(self, kind='mean', is_time=False):
        if self.is_fail:
            return None
        if self.is_group:
            if kind == 'best':
                return self.t_best if is_time else self.y_opt_best
            elif kind == 'mean':
                return self.t_mean if is_time else self.y_opt_mean
            elif kind == 'wrst':
                return self.t_wrst if is_time else self.y_opt_wrst
            else:
                raise ValueError('Invalid kind')
        else:
            return self.t if is_time else self.y_opt

    def get_opt_str(self, opts, pretty=True):
        res = []
        for id, v in opts.items():
            if isinstance(v, bool):
                res.append(f'{id}')
            else:
                res.append(f'{id}: {v}' if pretty else f'{id}-{v}')
        return ('; ' if pretty else '__').join(res)

    def init_from_bm(self, bm):
        self.bm_config = bm.bm_config
        self.op_config = bm.op_config

        self.bm_history = bm.bm_history
        self.op_history = bm.op_history

        self.bm_opts = bm.bm_opts
        self.op_opts = bm.op_opts

        self.d = bm.d
        self.n = bm.n

        self.bm_name = bm.bm_name
        self.op_name = bm.op_name

        self.bm_seed = bm.bm_seed
        self.op_seed = None

        self.is_max = bm.is_max
        self.is_min = bm.is_min
        self.y_opt = bm.y_opt

        self.t = bm.t

        self.y_list = bm.y_list

        self.is_fail = True if self.op_history['err'] else False

        self.is_init = True

    def init_from_data(self, data):
        self.bm_config = copy(data['config']['bm'])
        self.op_config = copy(data['config'])
        del self.op_config['bm']

        self.bm_history = copy(data['history']['bm'])
        self.op_history = copy(data['history'])
        del self.op_history['bm']

        self.bm_opts = copy(data['bm_opts'])
        self.op_opts = copy(data['op_opts'])

        self.d = self.bm_config['d']
        self.n = self.bm_config['n']

        self.bm_name = self.bm_config['name']
        self.op_name = self.op_config['name']

        self.bm_seed = self.bm_config['seed']
        self.op_seed = self.op_config['seed']

        self.is_max = self.bm_config.get('is_opti_max',
            'agent' in self.bm_name.lower())
        self.is_min = not self.is_max
        self.y_opt = self.bm_history['y_max' if self.is_max else 'y_min']

        self.t = self.bm_history['time_full']

        self.y_list = copy(self.bm_history['y_list'])

        self.is_fail = True if self.op_history['err'] else False

        self.is_init = True

    def info_table(self, prec=2, value_best=None, kind='mean', is_time=False,
                   prefix='        & ', fail='FAIL', best_cmd='fat',
                   postfix='', prefix_comment_inner='%       > ',
                   with_comment=False):
        form = '{:-10.' + str(prec) + 'e}'

        if self.is_fail:
            v = fail
        else:
            v = self.get(kind, is_time)
            v = form.format(v).strip()

        if value_best is not None:
            value_best = form.format(value_best).strip()

        text = ''

        if with_comment:
            text += prefix_comment_inner
            text += self.info_text_bm(with_prefix=False) + '\n'
            text += prefix_comment_inner
            text += self.info_text_op(with_prefix=False) + '\n'

        text += prefix

        if v == value_best and best_cmd:
            text += '\\' + best_cmd + '{'+ v + '}'
        else:
            text += v

        return text + postfix

    def info_text(self, prec=5, prec_time=1, len_max=21):
        text = ''
        text += '\n' + self.info_text_bm(len_max)
        text += '\n' + self.info_text_op(len_max)

        task = 'max' if self.is_max else 'min'

        form = '{:-14.' + str(prec) + 'e}'
        form_time = '{:-.' + str(prec_time) + 'e}'

        if self.is_group:
            if self.is_fail:
                text += '\n          ***FAIL***'
            else:
                text += '\n'
                text += '  > BEST >> '
                v = form.format(self.y_opt_best)
                text += f'{task}: {v}   '
                v = form_time.format(self.t_best)
                text += f'[time: {v}]'

                text += '\n'
                text += '  > MEAN >> '
                v = form.format(self.y_opt_mean)
                text += f'{task}: {v}   '
                v = form_time.format(self.t_mean)
                text += f'[time: {v}]'

                text += '\n'
                text += '  > WRST >> '
                v = form.format(self.y_opt_wrst)
                text += f'{task}: {v}   '
                v = form_time.format(self.t_wrst)
                text += f'[time: {v}]'

        else:
            text += '\n'
            text += '  >>>>>> '
            if self.is_fail:
                text += ' ***FAIL***'
            else:
                v = form.format(self.y_opt)
                text += f'{task}: {v}   '
                v = form_time.format(self.t)
                text += f'[time: {v}]'

        return text

    def info_text_bm(self, len_max=21, with_prefix=True):
        text = ''
        name = self.bm_name[:(len_max-1)]
        if with_prefix:
            pref = '- BM      > ' if self.is_group else '- BM   > '
        else:
            pref = ''
        opts = copy(self.bm_opts)
        text += pref + name + ' ' * max(0, len_max-len(name))
        text += ' [' + self.get_opt_str(opts, pretty=True) + ']'
        return text

    def info_text_op(self, len_max=21, with_prefix=True):
        text = ''
        name = self.op_name[:(len_max-1)]
        if with_prefix:
            pref = '- OPTI    > ' if self.is_group else '- OPTI > '
        else:
            pref = ''
        opts = copy(self.op_opts)
        if self.is_group:
            opts['SEEDS'] = len(self.op_seed_list)
            del opts['seed']
        text += pref + name + ' ' * max(0, len_max-len(name))
        text += ' [' + self.get_opt_str(opts, pretty=True) + ']'
        return text

    def is_better(self, value, kind='mean', is_time=False):
        if value is None:
            return True

        v = self.get(kind, is_time)
        if v is None:
            return False

        if is_time:
            return v < value
        else:
            return (value < v) if self.is_max else (value > v)

    def is_same(self, bm, skip_op_seed=True):
        if not self.is_init:
            return True

        bm1 = self
        bm2 = bm

        if bm1.bm_name !=bm2.bm_name:
            return False

        if bm1.op_name !=bm2.op_name:
            return False

        for id in bm1.bm_opts.keys():
            if not id in bm2.bm_opts:
                return False
            if bm1.bm_opts[id] != bm2.bm_opts[id]:
                return False

        for id in bm2.bm_opts.keys():
            if not id in bm1.bm_opts:
                return False
            if bm1.bm_opts[id] != bm2.bm_opts[id]:
                return False

        for id in bm1.op_opts.keys():
            if id == 'seed' and skip_op_seed:
                continue
            if not id in bm2.op_opts:
                return False
            if bm1.op_opts[id] != bm2.op_opts[id]:
                return False

        for id in bm2.op_opts.keys():
            if id == 'seed' and skip_op_seed:
                continue
            if not id in bm1.op_opts:
                return False
            if bm1.op_opts[id] != bm2.op_opts[id]:
                return False

        return True

    def values_to_one(self, y_lists, kind='mean'):
        res = []
        # lens = [len(y_list) for y_list in y_lists]
        # for k in range(max(lens)):
        for k in range(self.bm_config['budget_m']):
            vals = []
            for y_list in y_lists:
                l = len(y_list)
                vals.append(y_list[k] if l > k else y_list[l-1])
            if kind == 'best':
                res.append(np.max(vals))
            elif kind == 'mean':
                res.append(np.mean(vals))
            elif kind == 'wrst':
                res.append(np.min(vals))
            else:
                raise ValueError('Invalid kind')
        return res

    def values_to_opt(self, y_list):
        res = []
        opt = None
        for k, v in enumerate(y_list):
            if opt is None:
                opt = v
            elif self.is_max and v > opt:
                opt = v
            elif self.is_min and v < opt:
                opt = v
            res.append(opt)
        return res
